package unibo.cs.labss.editor.il.model.visitors;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import unibo.cs.labss.editor.il.model.Expression;
import unibo.cs.labss.editor.il.model.IILEntity;
import unibo.cs.labss.editor.il.model.IILVisitor;
import unibo.cs.labss.editor.il.model.Literal;
import unibo.cs.labss.editor.il.model.Statement;
import unibo.cs.labss.editor.il.model.impl.AbstractBinaryExpression;
import unibo.cs.labss.editor.il.model.impl.AdditionExpression;
import unibo.cs.labss.editor.il.model.impl.AndExpression;
import unibo.cs.labss.editor.il.model.impl.AssignmentStatement;
import unibo.cs.labss.editor.il.model.impl.BlockStatement;
import unibo.cs.labss.editor.il.model.impl.BooleanLiteral;
import unibo.cs.labss.editor.il.model.impl.DivisionExpression;
import unibo.cs.labss.editor.il.model.impl.EmptyBooleanExpression;
import unibo.cs.labss.editor.il.model.impl.EmptyExpression;
import unibo.cs.labss.editor.il.model.impl.EmptyIntegerExpression;
import unibo.cs.labss.editor.il.model.impl.EmptyStatement;
import unibo.cs.labss.editor.il.model.impl.EqualsExpression;
import unibo.cs.labss.editor.il.model.impl.GreaterThanExpression;
import unibo.cs.labss.editor.il.model.impl.IfElseStatement;
import unibo.cs.labss.editor.il.model.impl.IfStatement;
import unibo.cs.labss.editor.il.model.impl.IntegerLiteral;
import unibo.cs.labss.editor.il.model.impl.LessThanExpression;
import unibo.cs.labss.editor.il.model.impl.MultiplicationExpression;
import unibo.cs.labss.editor.il.model.impl.NotExpression;
import unibo.cs.labss.editor.il.model.impl.OrExpression;
import unibo.cs.labss.editor.il.model.impl.PrintStatement;
import unibo.cs.labss.editor.il.model.impl.Program;
import unibo.cs.labss.editor.il.model.impl.SubtractionExpression;
import unibo.cs.labss.editor.il.model.impl.Variable;
import unibo.cs.labss.editor.il.model.impl.WhileStatement;

/**
 * @author Riccardo Solmi
 */
public class InterpreterVisitor implements IILVisitor {
	private BindingManager bindingManager = BindingManager.create();
	private final PrintStream out;
	private Literal result = null;
	
	protected InterpreterVisitor(PrintStream out) {
		this.out = out;
	}

	public static void interpret(IILEntity program, PrintStream out) {
		program.accept(new InterpreterVisitor(out));
	}

	private final Literal evaluate(Expression exp) {
		exp.accept(this);
		return (Literal)result;
	}
	private final void execute(Statement stm) {
		stm.accept(this);
	}

	public void visitProgram(Program p) {
		p.getRootEntity().accept(this);
	}

	
	public void visitBlock(BlockStatement stm) {
		Iterator i = stm.iterator();
		while (i.hasNext())
			execute((Statement)i.next());
	}

	public void visitIf(IfStatement stm) {
		if (evaluate(stm.getExp()).booleanValue())
			execute(stm.getThenStm());
	}
	public void visitIfElse(IfElseStatement stm) {
		if (evaluate(stm.getExp()).booleanValue())
			execute(stm.getThenStm());
		else
			execute(stm.getElseStm());
	}

	public void visitAssignment(AssignmentStatement stm) {
		bindingManager.bind(stm.getId().getId(), evaluate(stm.getExp()));
//		stm.getId().setValue(evaluate(stm.getExp()));
	};

	public void visitWhile(WhileStatement stm) {
		while (evaluate(stm.getExp()).booleanValue())
			execute(stm.getStm());
	}

	public void visitPrint(PrintStatement stm) {
		out.println(evaluate(stm.getExp()));
	}

	public void visitBooleanLiteral(Literal exp) {
		result = exp;
	}
	public void visitIntegerLiteral(Literal exp) {
		result = exp;
	}
	public void visitVoidLiteral(Literal exp) {
		result = exp;
	}

	public void visitVariable(Variable exp) {
		result = bindingManager.getValue(exp.getId());
//		result = exp.getValue();
		if (result == null)
			throw new IllegalStateException("Variabile non inizializzata: "+exp.getId());
	}

	public void visitEmpty(EmptyStatement exp) {
	}

	public void visitEmpty(EmptyExpression exp) {
		throw new IllegalStateException("Espressione non definita");
	}
	public void visitEmptyBoolean(EmptyBooleanExpression exp) {
		throw new IllegalStateException("Espressione non definita");
	}
	public void visitEmptyInteger(EmptyIntegerExpression exp) {
		throw new IllegalStateException("Espressione non definita");
	}

	private final void setResult(int value) {
		result = IntegerLiteral.create(value);
	}
	private final int intValue(Expression exp) {
		return evaluate(exp).intValue();
	}

	static interface IxItoI {
		public int eval(int a, int b);
	}
	private final void evaluate(AbstractBinaryExpression exp, IxItoI op) {
		result = IntegerLiteral.create(op.eval(evaluate(exp.getExp1()).intValue(), evaluate(exp.getExp2()).intValue()));
	}

	public void visitAddition(AdditionExpression exp) {
		setResult(intValue(exp.getExp1()) + intValue(exp.getExp2()));
	}

	public void visitSubtraction(SubtractionExpression exp) {
		evaluate(exp, new IxItoI() {
			public int eval(int a, int b) {
				return a - b;
			}
		});
	}

	public void visitMultiplication(MultiplicationExpression exp) {
		result = IntegerLiteral.create(evaluate(exp.getExp1()).intValue() * evaluate(exp.getExp2()).intValue());
	}

	public void visitDivision(DivisionExpression exp) {
		result = IntegerLiteral.create(evaluate(exp.getExp1()).intValue() / evaluate(exp.getExp2()).intValue());
	}

	public void visitEquals(EqualsExpression exp) {
		result = BooleanLiteral.create(evaluate(exp.getExp1()).intValue() == evaluate(exp.getExp2()).intValue());
	}
	public void visitGreaterThan(GreaterThanExpression exp) {
		result = BooleanLiteral.create(evaluate(exp.getExp1()).intValue() > evaluate(exp.getExp2()).intValue());
	}
	public void visitLessThan(LessThanExpression exp) {
		result = BooleanLiteral.create(evaluate(exp.getExp1()).intValue() < evaluate(exp.getExp2()).intValue());
	}

	public void visitAnd(AndExpression exp) {
		result = BooleanLiteral.create(evaluate(exp.getExp1()).booleanValue() && evaluate(exp.getExp2()).booleanValue());
	}
	public void visitOr(OrExpression exp) {
		result = BooleanLiteral.create(evaluate(exp.getExp1()).booleanValue() || evaluate(exp.getExp2()).booleanValue());
	}

	public void visitNot(NotExpression exp) {
		result = BooleanLiteral.create(!evaluate(exp.getExp()).booleanValue());
	}

	
	public static class BindingManager {
		public static BindingManager create() {
			return new BindingManager() {
				public void exitScope() {
					throw new IllegalStateException("exitScope from top level");		
				}

				public boolean contains(String identifier) {
					return map.containsKey(identifier);
				}
				public Literal getValue(String identifier) {
					return (Literal)map.get(identifier);
				}
			};
		}

		private BindingManager enclosingScope;
		protected Map map = new HashMap();

		private BindingManager() {
		}
		private BindingManager(BindingManager enclosingScope, Map map) {
			this.enclosingScope = enclosingScope;
			this.map = map;
		}
		
		public void enterScope() {
			this.enclosingScope = new BindingManager(enclosingScope, map);
			this.map = new HashMap();
		}
		public void exitScope() {
			if (enclosingScope != null) {
				this.enclosingScope = enclosingScope.enclosingScope;
				this.map = enclosingScope.map;
			}		
		}

		public boolean contains(String identifier) {
			return map.containsKey(identifier) ? true : enclosingScope.contains(identifier);
		}
		public void bind(String identifier, Literal value) {
			map.put(identifier, value);
		}
		public Literal getValue(String identifier) {
			Literal value = (Literal)map.get(identifier);
			return (value != null) ? value : enclosingScope.getValue(identifier);
		}
	}
}
