package unibo.cs.labss.editor.il.model.visitors;

import java.io.PrintStream;
import java.util.Iterator;

import unibo.cs.labss.editor.il.model.Expression;
import unibo.cs.labss.editor.il.model.IILEntity;
import unibo.cs.labss.editor.il.model.IILVisitor;
import unibo.cs.labss.editor.il.model.Literal;
import unibo.cs.labss.editor.il.model.impl.AbstractBinaryExpression;
import unibo.cs.labss.editor.il.model.impl.AdditionExpression;
import unibo.cs.labss.editor.il.model.impl.AndExpression;
import unibo.cs.labss.editor.il.model.impl.AssignmentStatement;
import unibo.cs.labss.editor.il.model.impl.BlockStatement;
import unibo.cs.labss.editor.il.model.impl.DivisionExpression;
import unibo.cs.labss.editor.il.model.impl.EmptyBooleanExpression;
import unibo.cs.labss.editor.il.model.impl.EmptyExpression;
import unibo.cs.labss.editor.il.model.impl.EmptyIntegerExpression;
import unibo.cs.labss.editor.il.model.impl.EmptyStatement;
import unibo.cs.labss.editor.il.model.impl.EqualsExpression;
import unibo.cs.labss.editor.il.model.impl.GreaterThanExpression;
import unibo.cs.labss.editor.il.model.impl.IfElseStatement;
import unibo.cs.labss.editor.il.model.impl.IfStatement;
import unibo.cs.labss.editor.il.model.impl.LessThanExpression;
import unibo.cs.labss.editor.il.model.impl.MultiplicationExpression;
import unibo.cs.labss.editor.il.model.impl.NotExpression;
import unibo.cs.labss.editor.il.model.impl.OrExpression;
import unibo.cs.labss.editor.il.model.impl.PrintStatement;
import unibo.cs.labss.editor.il.model.impl.Program;
import unibo.cs.labss.editor.il.model.impl.SubtractionExpression;
import unibo.cs.labss.editor.il.model.impl.Variable;
import unibo.cs.labss.editor.il.model.impl.WhileStatement;

/**
 * @author Riccardo Solmi
 */
public class PrettyPrinterVisitor implements IILVisitor {
	protected final PrintStream out;
	private final String indentationString = "                                ";
	protected int indentation = 0;
	protected int indentationSize = 4;

	public PrettyPrinterVisitor(PrintStream out) {
		this.out = out;
	}

	public static void prettyPrint(IILEntity entity, PrintStream out) {
		entity.accept(new PrettyPrinterVisitor(out));
	}

	protected String indentationString() {
		if (indentation > 0) {
			if (indentation * indentationSize < indentationString.length())
				return indentationString.substring(0, indentation * indentationSize);
			else
				return indentationString;
		} else
			return "";
	}
	protected void printIndentation() {
		if (indentation > 0) {
			if (indentation * indentationSize < indentationString.length())
				out.print(indentationString.substring(0, indentation * indentationSize));
			else
				out.print(indentationString);
		}
	}

	protected void visitChild(Expression parent, Expression child) {
		if (parent.hasPrecedence(child)) {
			out.print("(");
			child.accept(this);
			out.print(")");
		} else
			child.accept(this);
	}
	protected void visitBinExp(AbstractBinaryExpression exp, String opStr) {
		visitChild(exp, exp.getExp1());
		out.print(opStr);
		visitChild(exp, exp.getExp2());
	}

	public void visitProgram(Program p) {
		out.println("public class Main {");
		indentation++;
		printIndentation();
		out.println("public static void main(String[] args) {");
		indentation++;
		p.getRootEntity().accept(this);
		indentation--;
		printIndentation();
		out.println("}");
		indentation--;
		printIndentation();
		out.println("}");
	}

	public void visitBlock(BlockStatement stm) {
		Iterator i = stm.iterator();
		while (i.hasNext())
			((IILEntity)i.next()).accept(this);
	}

	public void visitIf(IfStatement stm) {
		printIndentation();
		out.print("if (");
		stm.getExp().accept(this);
		indentation++;
		if (stm.getThenStm() instanceof BlockStatement) {
			out.println(") {");
			stm.getThenStm().accept(this);
			out.println("}");
		} else {
			out.println(")");
			stm.getThenStm().accept(this);
		}
		indentation--;
	}
	public void visitIfElse(IfElseStatement stm) {
		printIndentation();
		out.print("if (");
		stm.getExp().accept(this);
		indentation++;
		if (stm.getThenStm() instanceof BlockStatement) {
			out.println(") {");
			stm.getThenStm().accept(this);
			out.println("}");
		} else {
			out.println(")");
			stm.getThenStm().accept(this);
		}
		indentation--;

		printIndentation();
		indentation++;
		if (stm.getElseStm() instanceof BlockStatement) {
			out.println("else {");
			stm.getElseStm().accept(this);
			out.println("}");
		} else {
			out.println("else");
			stm.getElseStm().accept(this);
		}
		indentation--;
	}

	public void visitAssignment(AssignmentStatement stm) {
		printIndentation();
		stm.getId().accept(this);
		out.print(" = ");
		stm.getExp().accept(this);
		out.println(";");
	}

	public void visitWhile(WhileStatement stm) {
		printIndentation();
		out.print("while (");
		stm.getExp().accept(this);
		indentation++;
		if (stm.getStm() instanceof BlockStatement) {
			out.println(") {");
			stm.getStm().accept(this);
			indentation--;
			printIndentation();
			out.println("}");
		} else {
			out.println(")");
			stm.getStm().accept(this);
			indentation--;
		}
	}

	public void visitPrint(PrintStatement stm) {
		printIndentation();
		out.print("System.out.println(");
		stm.getExp().accept(this);
		out.println(");");
	}

	public void visitBooleanLiteral(Literal exp) {
		out.print(exp);
	}
	public void visitIntegerLiteral(Literal exp) {
		out.print(exp);
	}
	public void visitVoidLiteral(Literal exp) {
		out.print(exp);
	}

	public void visitVariable(Variable exp) {
		out.print(exp.getId());
	}

	public void visitEmpty(EmptyStatement exp) {
	}
	public void visitEmpty(EmptyExpression exp) {
	}
	public void visitEmptyBoolean(EmptyBooleanExpression exp) {
	}
	public void visitEmptyInteger(EmptyIntegerExpression exp) {
	}

	public void visitAddition(AdditionExpression exp) {
		visitBinExp(exp, " + ");
	}
	public void visitSubtraction(SubtractionExpression exp) {
		visitBinExp(exp, " - ");
	}
	public void visitMultiplication(MultiplicationExpression exp) {
		visitBinExp(exp, " * ");
	}
	public void visitDivision(DivisionExpression exp) {
		visitBinExp(exp, " / ");
	}
	public void visitEquals(EqualsExpression exp) {
		visitBinExp(exp, " == ");
	}
	public void visitGreaterThan(GreaterThanExpression exp) {
		visitBinExp(exp, " > ");
	}
	public void visitLessThan(LessThanExpression exp) {
		visitBinExp(exp, " < ");
	}
	public void visitAnd(AndExpression exp) {
		visitBinExp(exp, " && ");
	}
	public void visitOr(OrExpression exp) {
		visitBinExp(exp, " || ");
	}
	public void visitNot(NotExpression exp) {
		out.print("!");
		visitChild(exp, exp.getExp());
	}
}
